import json
from datetime import datetime, date

from memorious.core import connect_redis


class Base(object):
    conn = connect_redis()


def unpack_int(value):
    try:
        return int(value)
    except Exception:
        return 0


def pack_datetime(value):
        if value is not None:
            return str(value)


def pack_now():
    return pack_datetime(datetime.utcnow())


def unpack_datetime(value):
    if value is not None:
        return datetime.strptime(value, "%Y-%m-%d %H:%M:%S.%f")


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, bytes):
            return obj.decode('utf-8')
        if isinstance(obj, set):
            return [o for o in obj]
        return json.JSONEncoder.default(self, obj)


def dump_json(data):
    if data is None:
        return
    return JSONEncoder().encode(data)


def load_json(encoded):
    if encoded is None:
        return
    return json.loads(encoded)



import os
import json
import shutil
from normality import safe_filename

from memorious import settings


def _get_directory_path(context):
    """Get the storage path fro the output."""
    path = os.path.join(settings.BASE_PATH, 'store')
    path = context.params.get('path', path)
    path = os.path.join(path, context.crawler.name)
    path = os.path.abspath(os.path.expandvars(path))
    try:
        os.makedirs(path)
    except Exception:
        pass
    return path


def directory(context, data):
    """Store the collected files to a given directory."""
    with context.http.rehash(data) as result:
        if not result.ok:
            return

        content_hash = data.get('content_hash')
        if content_hash is None:
            context.emit_warning("No content hash in data.")
            return

        path = _get_directory_path(context)
        file_name = data.get('file_name', result.file_name)
        file_name = safe_filename(file_name, default='raw')
        file_name = '%s.%s' % (content_hash, file_name)
        data['_file_name'] = file_name
        file_path = os.path.join(path, file_name)
        if not os.path.exists(file_path):
            shutil.copyfile(result.file_path, file_path)

        context.log.info("Store [directory]: %s", file_name)

        meta_path = os.path.join(path, '%s.json' % content_hash)
        with open(meta_path, 'w') as fh:
            json.dump(data, fh)



from importlib import import_module

from memorious.model import CrawlerState


class CrawlerStage(object):
    """A single step in a data processing crawler."""

    def __init__(self, crawler, name, config):
        self.crawler = crawler
        self.name = name
        self.config = config
        self.method_name = config.get('method')
        self.params = config.get('params') or {}
        self.handlers = config.get('handle') or {}
        self.rate_limit = None
        if 'rate_limit' in config:
            self.rate_limit = float(config.get('rate_limit'))

    @property
    def method(self):
        method = self.method_name
        package = 'memorious.operations'
        if ':' in method:
            package, method = method.rsplit(':', 1)
        module = import_module(package)
        return getattr(module, method)

    @property
    def op_count(self):
        """Total operations performed for this stage"""
        return CrawlerState.op_count(self.crawler, self)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<CrawlerStage(%r, %s)>' % (self.crawler, self.name)




WEB = [
    'text/html',
    'text/plain',
    'text/xml',
    'application/xml',
]

IMAGES = [
    'image/jpeg',
    'image/bmp',
    'image/png',
    'image/tiff',
    'image/gif',
    'application/postscript',
    'image/vnd.dxf',
    'image/svg+xml',
    'image/x-pict',
    'image/x-ms-bmp'
]

MEDIA = [
    'audio/mpeg',
    'video/x-flv',
    'video/mp4',
    'video/mp2t',
    'video/3gpp',
    'video/quicktime',
    'video/x-msvideo',
    'video/x-ms-wmv',
]

DOCUMENTS = [
    'application/vnd.ms-excel',
    'application/msword',
    'application/pdf',
    '.pdf',  # seen in the wild (TM)
    'application/vnd.ms-powerpoint',
    'text/richtext',
    'application/vnd.oasis.opendocument.text',
    'application/rtf',
    'application/vnd.oasis.opendocument.spreadsheet',
    'application/x-rtf',
    'application/vnd.oasis.opendocument.graphics',
    'application/vnd.oasis.opendocument.presentation',
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet', # noqa
    'application/vnd.openxmlformats-officedocument.wordprocessingml.document', # noqa
    'application/vnd.openxmlformats-officedocument.presentationml.presentation' # noqa
]

ARCHIVES = [
    'application/zip',
    'application/x-rar-compressed',
    'application/x-tar',
    'application/x-gzip',
    'application/x-7z-compressed'
]

ASSETS = [
    'text/css',
    'application/javascript',
    'application/json',
    'image/x-icon',
    'application/rss+xml',
    'application/atom+xml',
    'application/opensearchdescription+xml',
]

GROUPS = {
    'web': WEB,
    'images': IMAGES,
    'media': MEDIA,
    'documents': DOCUMENTS,
    'archives': ARCHIVES,
    'assets': ASSETS
}

KNOWN = WEB + IMAGES + MEDIA + DOCUMENTS + ARCHIVES + ASSETS
NON_HTML = [t for t in KNOWN if t not in WEB]



import queue
import random
import logging
import threading
from datetime import datetime

from memorious import settings
from memorious.logic.rate_limit import rate_limiter, RateLimitException
from memorious.logic.context import Context
from memorious.model.queue import Queue

log = logging.getLogger(__name__)


class TaskRunner(object):
    """A long running task runner that uses Redis as a task queue"""

    @classmethod
    def execute(cls, stage, state, data, next_allowed_exec_time=None):
        """Execute the operation, rate limiting allowing."""
        now = datetime.utcnow()
        if next_allowed_exec_time and now < next_allowed_exec_time:
            # task not allowed to run yet; put it back in the queue
            Queue.queue(stage, state, data, delay=next_allowed_exec_time)
            return

        context = Context.from_state(state, stage)
        if context.crawler.disabled:
            return

        if context.stage.rate_limit:
            try:
                with rate_limiter(context):
                    context.execute(data)
                    return
            except RateLimitException:
                delay = max(1, 1.0/context.stage.rate_limit)
                delay = random.randint(1, int(delay))
                context.log.info(
                    "Rate limit exceeded, delaying %d sec.", delay
                )
                Queue.queue(stage, state, data, delay=delay)

        context.execute(data)

    @classmethod
    def process(cls, q):
        while True:
            item = q.get()
            if item is None:
                q.task_done()
                return
            try:
                cls.execute(*item)
            except Exception:
                log.exception("Task failed to execute:")
            q.task_done()

    @classmethod
    def run(cls):
        if settings.DEBUG:
            log.info("DEBUG mode. Disabling task threading.")
            for task in Queue.tasks():
                cls.execute(*task)
            return

        log.info("Processing queue (%s threads)", settings.THREADS)
        q = queue.Queue(maxsize=settings.THREADS)
        threads = []
        for i in range(settings.THREADS):
            t = threading.Thread(target=cls.process, args=(q,))
            t.daemon = True
            t.start()
            threads.append(t)

        while True:
            # This is to handle fakeredis. It terminates the iterator
            # in Queue.tasks() when it "temporarily" runs out of tasks
            # during the execution of the initial task. So we have to
            # restart Queue.tasks() after q.join() to make sure that
            # no child tasks were queued.
            tasks_processed = 0
            for item in Queue.tasks():
                tasks_processed += 1
                q.put(item)

            # block until all tasks are done
            q.join()

            if tasks_processed == 0:
                break

        # stop workers
        for i in range(settings.THREADS):
            q.put(None)
        for t in threads:
            t.join()



import logging
from six.moves.urllib.parse import urljoin
from banal import ensure_list
from urlnormalizer import normalize_url
from normality import collapse_spaces

from memorious.helpers.rule import Rule
from memorious.helpers.dates import iso_date
from memorious.util import make_key


log = logging.getLogger(__name__)
URL_TAGS = [('.//a', 'href'),
            ('.//img', 'src'),
            ('.//link', 'href'),
            ('.//iframe', 'src')]


def parse_html(context, data, result):
    context.log.info('Parse: %r', result.url)

    title = result.html.findtext('.//title')
    if title is not None and 'title' not in data:
        data['title'] = title

    include = context.params.get('include_paths')
    if include is None:
        roots = [result.html]
    else:
        roots = []
        for path in include:
            roots = roots + result.html.findall(path)

    seen = set()
    for root in roots:
        for tag_query, attr_name in URL_TAGS:
            for element in root.findall(tag_query):
                attr = element.get(attr_name)
                if attr is None:
                    continue

                try:
                    url = normalize_url(urljoin(result.url, attr))
                except Exception:
                    log.warning('Invalid URL: %r', attr)
                    continue

                if url is None or url in seen:
                    continue
                seen.add(url)

                tag = make_key(context.run_id, url)
                if context.check_tag(tag):
                    continue
                context.set_tag(tag, None)
                data = {'url': url}
                # Option to set the document title from the link text.
                if context.get('link_title', False):
                    data['title'] = collapse_spaces(element.text_content())
                elif element.get('title'):
                    data['title'] = collapse_spaces(element.get('title'))

                context.http.session.headers['Referer'] = url
                context.emit(rule='fetch', data=data)


def parse_for_metadata(context, data, html):
    meta = context.params.get('meta', {})
    meta_date = context.params.get('meta_date', {})

    meta_paths = meta
    meta_paths.update(meta_date)

    for key, xpaths in meta_paths.items():
        for xpath in ensure_list(xpaths):
            element = html.find(xpath)
            if element is None:
                continue
            value = collapse_spaces(element.text_content())
            if key in meta_date:
                value = iso_date(value)
            if value is not None:
                data[key] = value
            break

    return data


def parse(context, data):
    with context.http.rehash(data) as result:
        if result.html is not None:
            parse_html(context, data, result)

            # Get extra metadata from the DOM
            parse_for_metadata(context, data, result.html)

        rules = context.params.get('store') or {'match_all': {}}
        if Rule.get_rule(rules).apply(result):
            context.emit(rule='store', data=data)



import math
import logging

from memorious import settings
from memorious.core import manager
from memorious.model import Event

log = logging.getLogger(__name__)


def global_stats():
    """Stats visible on each page of the UI."""
    stats = {
        'version': settings.VERSION,
        'num_crawlers': len(manager)
    }
    return stats


def crawlers_index():
    """Generate a list of all crawlers, sorted alphabetically, with op
    counts."""
    crawlers = []
    for crawler in manager:
        data = Event.get_counts(crawler)
        data['last_active'] = crawler.last_run
        data['total_ops'] = crawler.op_count
        data['running'] = crawler.is_running
        data['crawler'] = crawler
        crawlers.append(data)
    return crawlers


def get_crawler(name):
    return manager.get(name)


def crawler_stages(crawler):
    """See the number of executions of each stage."""
    stages = []
    for stage in crawler:
        data = Event.get_stage_counts(crawler, stage)
        data['total_ops'] = stage.op_count
        data['stage'] = stage
        stages.append(data)
    return stages


def crawler_events(crawler, level=None, stage_name=None,
                   run_id=None, page=1, per_page=15):
    start = (max(1, page) - 1) * per_page
    end = start + per_page

    if stage_name:
        events = Event.get_stage_events(crawler, stage_name, start, end, level)
    elif run_id:
        events = Event.get_run_events(crawler, run_id, start, end, level)
    else:
        events = Event.get_crawler_events(crawler, start, end, level)
    total = len(events)

    return {
        'page': page,
        'per_page': per_page,
        'pages': int(math.ceil((float(total) / per_page))),
        'total': total,
        'results': events
    }


def crawler_runs(crawler):
    runs = list(crawler.runs)
    for run in runs:
        run.update(Event.get_run_counts(crawler, run['run_id']))
    return runs



from urllib.parse import urlencode

from flask import Flask, jsonify, request
from flask import render_template, abort
from babel.numbers import format_number
from babel.dates import format_date, format_datetime
from raven.contrib.flask import Sentry

from memorious.ui.reporting import (
    crawlers_index, global_stats, get_crawler,
    crawler_stages, crawler_events,
    crawler_runs
)

app = Flask(__name__)
sentry = Sentry()


# @app.after_request
# def cleanup(response):
#     return response


@app.template_filter('number')
def number_filter(s, default=''):
    if s is None or s == 0 or not len(str(s)):
        return default
    return format_number(s, locale='en_GB')


@app.template_filter('datetime')
def datetime_filter(s):
    if s is None or s == 0 or not len(str(s)):
        return ''
    return format_datetime(s, locale='en_GB', format='short')


@app.template_filter('date')
def date_filter(s):
    if s is None or s == 0 or not len(str(s)):
        return ''
    return format_date(s, locale='en_GB', format='short')


def state_change(name, value):
    state = [(name, value)]
    for aname, avalue in request.args.items():
        if aname != name:
            state.append((aname, avalue))
    return '?' + urlencode(state)


@app.context_processor
def context():
    context = global_stats()
    context['state_change'] = state_change
    return context


@app.route('/')
def index():
    crawlers = crawlers_index()
    return render_template('index.html', crawlers=crawlers)


@app.route('/crawlers/<name>')
def crawler(name):
    crawler = get_crawler(name)
    if crawler is None:
        abort(404)
    stages = crawler_stages(crawler)
    runs = crawler_runs(crawler)
    return render_template('crawler.html',
                           crawler=crawler,
                           stages=stages, runs=runs)


@app.route('/crawlers/<name>/events')
def events(name):
    crawler = get_crawler(name)
    if crawler is None:
        abort(404)
    events = crawler_events(crawler,
                            page=int(request.args.get('page', 1)),
                            run_id=request.args.get('run_id'),
                            level=request.args.get('level'),
                            stage_name=request.args.get('stage_name'))
    return render_template('events.html',
                           crawler=crawler,
                           events=events)


@app.route('/crawlers/<name>/config')
def config(name):
    crawler = get_crawler(name)
    if crawler is None:
        abort(404)
    return render_template('config.html', crawler=crawler)


@app.route('/invoke/<crawler>/<action>', methods=['POST'])
def invoke(crawler, action):
    crawler = get_crawler(crawler)
    if crawler is None:
        abort(400)
    if action == 'run':
        crawler.run()
        return jsonify({'status': 'ok'})
    if action == 'cancel':
        crawler.cancel()
        return jsonify({'status': 'ok'})
    if action == 'flush':
        crawler.flush()
        return jsonify({'status': 'ok'})
    if action == 'flush-events':
        crawler.flush_events()
        return jsonify({'status': 'ok'})
    else:
        abort(400)



import logging

from memorious.model.common import Base, unpack_int, unpack_datetime, pack_now
from memorious.util import make_key

log = logging.getLogger(__name__)


class CrawlerState(Base):
    """The current state of a running crawler instance"""

    @classmethod
    def last_run(cls, crawler):
        last_run = cls.conn.get(make_key(crawler, "last_run"))
        return unpack_datetime(last_run)

    @classmethod
    def op_count(cls, crawler, stage=None):
        """Total operations performed for this crawler"""
        if stage:
            total_ops = cls.conn.get(make_key(crawler, stage))
        else:
            total_ops = cls.conn.get(make_key(crawler, "total_ops"))
        return unpack_int(total_ops)

    @classmethod
    def runs(cls, crawler):
        for run_id in cls.conn.lrange(make_key(crawler, "runs_list"), 0, -1):
            start = cls.conn.get(make_key("run", run_id, "start"))
            end = cls.conn.get(make_key("run", run_id, "end"))
            total_ops = cls.conn.get(make_key("run", run_id, "total_ops"))
            yield {
                'run_id': run_id,
                'total_ops': unpack_int(total_ops),
                'start': unpack_datetime(start),
                'end': unpack_datetime(end)
            }

    @classmethod
    def record_operation_start(cls, crawler, stage):
        cls.conn.incr(make_key(crawler, stage))
        cls.conn.incr(make_key(crawler, "total_ops"))
        cls.conn.set(make_key(crawler, "last_run"), pack_now())

    @classmethod
    def record_operation_end(cls, crawler):
        cls.conn.set(make_key(crawler, "last_run"), pack_now())

    @classmethod
    def latest_runid(cls, crawler):
        return cls.conn.lindex(make_key(crawler, "runs_list"), 0)

    @classmethod
    def flush(cls, crawler):
        cls.conn.delete(make_key(crawler, "total_ops"))
        cls.conn.delete(make_key(crawler, "last_run"))
        for stage in crawler.stages:
            cls.conn.delete(make_key(crawler, stage))
